"""
RETURNN frontend (returnn.frontend) tests
"""

from __future__ import annotations
from typing import Tuple
import _setup_test_env  # noqa
import returnn.frontend as rf
from returnn.tensor import Tensor, Dim, TensorDict, batch_dim
from rf_utils import run_model


def test_cond():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    out_dim = Dim(13, name="out")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __init__(self):
            super().__init__()
            self.linear_true = rf.Linear(in_dim, out_dim)
            self.linear_false = rf.Linear(in_dim, out_dim)

        def __call__(self, x: Tensor) -> Tensor:
            return rf.cond(
                pred=batch_dim.get_dim_value_tensor() % 2 == 0,
                true_fn=lambda: self.linear_true(x),
                false_fn=lambda: self.linear_false(x),
            )

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"])
        out.mark_as_default_output(shape=(batch_dim, time_dim, out_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step, test_single_batch_entry=False)


def test_cond_via_time_even():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    out_dim = Dim(13, name="out")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __init__(self):
            super().__init__()
            self.linear_true = rf.Linear(in_dim, out_dim)
            self.linear_false = rf.Linear(in_dim, out_dim)

        def __call__(self, x: Tensor) -> Tensor:
            return rf.cond(
                pred=time_dim.get_dim_value_tensor() % 2 == 0,
                true_fn=lambda: self.linear_true(x),
                false_fn=lambda: self.linear_false(x),
            )

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"])
        out.mark_as_default_output(shape=(batch_dim, time_dim, out_dim))

    run_model(
        extern_data,
        lambda *, epoch, step: _Net(),
        _forward_step,
        dyn_dim_max_sizes={time_dim: 5},
        test_single_batch_entry=False,
    )
    run_model(
        extern_data,
        lambda *, epoch, step: _Net(),
        _forward_step,
        dyn_dim_max_sizes={time_dim: 6},
        test_single_batch_entry=False,
    )


def test_cond_shared_params():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    out_dim = Dim(13, name="out")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __init__(self):
            super().__init__()
            self.linear = rf.Linear(in_dim, out_dim)

        def __call__(self, x: Tensor) -> Tensor:
            return rf.cond(
                pred=time_dim.get_dim_value_tensor() % 2 == 0,
                true_fn=lambda: self.linear(x),
                false_fn=lambda: self.linear(x * 2.0),
            )

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"])
        out.mark_as_default_output(shape=(batch_dim, time_dim, out_dim))

    run_model(
        extern_data,
        lambda *, epoch, step: _Net(),
        _forward_step,
        dyn_dim_max_sizes={time_dim: 5},
        test_single_batch_entry=False,
    )
    run_model(
        extern_data,
        lambda *, epoch, step: _Net(),
        _forward_step,
        dyn_dim_max_sizes={time_dim: 6},
        test_single_batch_entry=False,
    )


def test_cond_twice_shared_params():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    out_dim = Dim(13, name="out")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __init__(self):
            super().__init__()
            self.pre_linear = rf.Linear(in_dim, out_dim)
            self.linear_true = rf.Linear(out_dim, out_dim)
            self.linear_false = rf.Linear(out_dim, out_dim)

        def __call__(self, x: Tensor) -> Tensor:
            x = self.pre_linear(x)
            x = rf.cond(
                pred=time_dim.get_dim_value_tensor() % 2 == 0,
                true_fn=lambda: self.linear_true(x),
                false_fn=lambda: self.linear_false(x),
            )
            x = rf.cond(
                pred=time_dim.get_dim_value_tensor() % 2 == 1,
                true_fn=lambda: self.linear_true(x),
                false_fn=lambda: self.linear_false(x),
            )
            return x

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"])
        out.mark_as_default_output(shape=(batch_dim, time_dim, out_dim))

    run_model(
        extern_data,
        lambda *, epoch, step: _Net(),
        _forward_step,
        dyn_dim_max_sizes={time_dim: 5},
        test_single_batch_entry=False,
    )
    run_model(
        extern_data,
        lambda *, epoch, step: _Net(),
        _forward_step,
        dyn_dim_max_sizes={time_dim: 6},
        test_single_batch_entry=False,
    )


def test_cond_param_assign():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __init__(self):
            super().__init__()
            self.param = rf.Parameter(dims=(), dtype="int32")
            self.param.initial = 2

        def __call__(self, x: Tensor) -> Tensor:
            # No extra care should be needed for graph-based backends.
            rf.cond(
                pred=time_dim.get_dim_value_tensor() % 2 == 0,
                true_fn=lambda: self.param.assign_add(3),
                false_fn=lambda: None,
            )
            return self.param

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"])
        out.mark_as_default_output(shape=())

    out1 = run_model(
        extern_data,
        lambda *, epoch, step: _Net(),
        _forward_step,
        dyn_dim_max_sizes={time_dim: 5},
        test_single_batch_entry=False,
    )
    out2 = run_model(
        extern_data,
        lambda *, epoch, step: _Net(),
        _forward_step,
        dyn_dim_max_sizes={time_dim: 6},
        test_single_batch_entry=False,
    )
    assert out1["output"].raw_tensor == 2
    assert out2["output"].raw_tensor == 5


def test_cond_param_assign2():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __init__(self):
            super().__init__()
            self.param = rf.Parameter(dims=(), dtype="int32")
            self.param.initial = 2

        def __call__(self, x: Tensor) -> Tensor:
            # No extra care should be needed for graph-based backends.
            rf.cond(
                pred=time_dim.get_dim_value_tensor() % 2 == 0,
                true_fn=lambda: self.param.assign_add(3),
                false_fn=lambda: self.param.assign_add(7),
            )
            return self.param

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"])
        out.mark_as_default_output(shape=())

    out1 = run_model(
        extern_data,
        lambda *, epoch, step: _Net(),
        _forward_step,
        dyn_dim_max_sizes={time_dim: 5},
        test_single_batch_entry=False,
    )
    out2 = run_model(
        extern_data,
        lambda *, epoch, step: _Net(),
        _forward_step,
        dyn_dim_max_sizes={time_dim: 6},
        test_single_batch_entry=False,
    )
    assert out1["output"].raw_tensor == 9
    assert out2["output"].raw_tensor == 5


def test_cond_param_assign3():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __init__(self):
            super().__init__()
            self.param = rf.Parameter(dims=(), dtype="int32")
            self.param.initial = 2

        def __call__(self, x: Tensor) -> Tuple[Tensor, Tensor]:
            # No extra care should be needed for graph-based backends.
            return (
                rf.cond(
                    pred=time_dim.get_dim_value_tensor() % 2 == 0,
                    true_fn=lambda: (self.param.assign_add(3), rf.convert_to_tensor(42))[-1],
                    false_fn=lambda: self.param * 3,
                ),
                self.param,
            )

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out, param = model(extern_data["data"])
        out.mark_as_default_output(shape=())
        param.mark_as_output(shape=(), name="param")

    out1 = run_model(
        extern_data,
        lambda *, epoch, step: _Net(),
        _forward_step,
        dyn_dim_max_sizes={time_dim: 5},
        test_single_batch_entry=False,
    )
    out2 = run_model(
        extern_data,
        lambda *, epoch, step: _Net(),
        _forward_step,
        dyn_dim_max_sizes={time_dim: 6},
        test_single_batch_entry=False,
    )
    assert out1["output"].raw_tensor == 6 and out1["param"].raw_tensor == 2
    assert out2["output"].raw_tensor == 42 and out2["param"].raw_tensor == 5


# TODO some more from RETURNN-common, like:
#  - test_cond_random
#  - test_cond_new_axis
#  - test_cond_dim (https://github.com/rwth-i6/returnn/pull/1262)
#  - test_cond_multiple_outputs



"""
This is the Sprint interface implementation,
concretely for Sprint PythonControl, Sprint PythonSegmentOrder and Sprint SprintNnPythonLayer.
For reference, in Sprint code, see:
 * src/Nn/PythonControl.cc
 * src/Tools/NnTrainer/python_control_demo.py
This interface will behave similar to SprintExternInterface.
See SprintErrorSignals for the other end.
It can also be used as a PythonSegmentOrdering interface.
It also supports SprintNnPythonLayer.
"""

from __future__ import annotations

import os
import numpy
import typing
from threading import Condition

import returnn.__main__ as rnn
import returnn.util.task_system as task_system
import returnn.util.debug as debug
from returnn.util.task_system import numpy_set_unused
from returnn.util.basic import to_bool, long
import returnn.util.basic as util

InitTypes = set()
Verbose = False  # disables all per-segment log messages
Quiet = False  # disables all but error messages

_orig_print = print


# noinspection PyShadowingBuiltins
def print(*args, **kwargs):
    """
    ``print`` replacement.

    :param args:
    :param kwargs:
    """
    if not Quiet:
        _orig_print(*args, **kwargs)


print("RETURNN SprintControl[pid %i] Python module load" % os.getpid())

rnn.init_better_exchook()
debug.init_faulthandler(sigusr1_chain=True)  # Sprint also handles SIGUSR1.


# Start Sprint PythonControl interface. {


def init(name, reference, config, sprint_unit=None, version_number=None, callback=None, **kwargs):
    """
    This will be called by Sprint PythonControl.
    But we also call it ourselves e.g. in getSegmentList() and SprintNnPythonLayer.
    In this specific module, we expect that there is "c2p_fd" and "p2c_fd" in the config string
    to communicate with the parent process, which is usually handled by SprintErrorSignals.

    :param str name: this specifies the caller. e.g. "Sprint.PythonControl"
    :param reference: this is any object to identify the specific instance of the caller, if there are multiple.
    :param str config: this will be passed over from Sprint. you can configure that via --*.pymod-config.
    :param str sprint_unit: if this is called by Sprint PythonControl, this will specify which specific part
      of Sprint is using this PythonControl, because there can be multiple parts.
      E.g. there is "FeedForwardTrainer", "SegmentwiseNnTrainer" and "NnTrainer.pythonControl".
    :param int|None version_number: if this is called by Sprint PythonControl, this will set the version number.
      only newer Sprint versions will set this.
    :param function|None callback: if this is called by Sprint PythonControl, this might provide a callback.
      Only newer Sprint versions will provide this to init().
      This callback can be used for many different actions.
      It's supposed to be called like callback(action, **other_args), where action is a string.
      See Sprint PythonControl code about the possible actions and arguments.
    :param kwargs: all remaining args are specific for each caller.
    """

    config = config.split(",")
    config = {key: value for (key, value) in [s.split(":", 1) for s in config if s]}

    global Quiet
    if to_bool(config.get("quiet", False)):
        Quiet = True

    print(
        (
            "RETURNN SprintControl[pid %i] init: "
            "name=%r, sprint_unit=%r, version_number=%r, callback=%r, ref=%r, config=%r, kwargs=%r"
        )
        % (os.getpid(), name, sprint_unit, version_number, callback, reference, config, kwargs)
    )
    InitTypes.add(name)

    global Verbose
    if to_bool(config.get("verbose", False)):
        Verbose = True

    if (
        to_bool(config.get("EnableAutoNumpySharedMemPickling", False))
        and not task_system.SharedMemNumpyConfig["enabled"]
    ):
        task_system.SharedMemNumpyConfig["enabled"] = True
        print("RETURNN SprintControl[pid %i] EnableAutoNumpySharedMemPickling = True" % (os.getpid(),))

    # Remaining Sprint interface is in this PythonControl instance.
    return PythonControl.create(
        c2p_fd=int(config["c2p_fd"]),
        p2c_fd=int(config["p2c_fd"]),
        name=name,
        reference=reference,
        config=config,
        sprint_unit=sprint_unit,
        version_number=version_number,
        min_version_number=int(config["minPythonControlVersion"]),
        callback=callback,
        **kwargs,
    )


# End Sprint PythonControl interface. }


# Start Sprint PythonSegmentOrder interface. {


# Keep names for compatibility.
# noinspection PyPep8Naming,PyUnusedLocal
def getSegmentList(corpusName, segmentList, config, **kwargs):
    """
    Sprint will directly call this function.
    """
    print("RETURNN SprintControl[pid %i] getSegmentList: corpus=%r, config=%r" % (os.getpid(), corpusName, config))

    # If we were not initialized via PythonControl interface, this will initialize us
    # and setup the communication channel (PythonControl).
    init(name="RETURNN.PythonSegmentOrder", reference=corpusName, config=config)
    PythonControl.instance.check_control_loop_running()
    for segment_name in PythonControl.instance.segment_list_iterator():
        if isinstance(segment_name, bytes):
            yield segment_name.decode("utf-8")
        else:
            yield segment_name


# End Sprint PythonSegmentOrder interface. }


# Start SprintNnPythonLayer. {


class SprintNnPythonLayer:
    """
    Sprint will directly call this class, i.e. create an instance of it.
    It implements the Sprint NN PythonLayer interface.
    """

    def __init__(self, config, **kwargs):
        print("RETURNN SprintControl[pid %i] SprintNnPythonLayer.__init__: %r, %r" % (os.getpid(), config, kwargs))
        # If we were not initialized via PythonControl interface, this will initialize us
        # and setup the communication channel (PythonControl).
        init(name="RETURNN.SprintNnPythonLayer", reference=self, config=config)
        self.input_size = None
        self.output_size = None

    # noinspection PyMethodMayBeStatic
    def finalize(self):
        """
        Called by Sprint at exit.
        """
        print("RETURNN SprintControl[pid %i] SprintNnPythonLayer.finalize" % (os.getpid(),))

    # noinspection PyPep8Naming
    def setInputDimension(self, stream, size):
        """
        :param int stream:
        :param int size:
        """
        print(
            "RETURNN SprintControl[pid %i] SprintNnPythonLayer.setInputDimension: stream=%r, size=%r"
            % (os.getpid(), stream, size)
        )
        assert stream == 0, "we only support a single input stream (for now)"
        self.input_size = size

    # noinspection PyPep8Naming
    def setOutputDimension(self, size):
        """
        :param int size:
        """
        print("RETURNN SprintControl[pid %i] SprintNnPythonLayer.setOutputDimension: %r" % (os.getpid(), size))
        self.output_size = size

    # noinspection PyPep8Naming,PyMethodMayBeStatic
    def initializeNetworkParameters(self):
        """
        Called by Sprint for param init.
        """
        print("RETURNN SprintControl[pid %i] SprintNnPythonLayer.initializeNetworkParameters" % (os.getpid(),))
        # Just ignore.

    # noinspection PyPep8Naming,PyMethodMayBeStatic
    def loadNetworkParameters(self, filename):
        """
        :param str filename:
        """
        print("RETURNN SprintControl[pid %i] SprintNnPythonLayer.loadNetworkParameters: %r" % (os.getpid(), filename))
        # Just ignore.

    # noinspection PyPep8Naming,PyMethodMayBeStatic
    def saveNetworkParameters(self, filename):
        """
        :param str filename:
        """
        print("RETURNN SprintControl[pid %i] SprintNnPythonLayer.saveNetworkParameters: %r" % (os.getpid(), filename))
        # Just ignore.

    # noinspection PyPep8Naming,PyMethodMayBeStatic
    def isTrainable(self):
        """
        :rtype: bool
        """
        # Always trainable.
        return True

    # noinspection PyPep8Naming,PyMethodMayBeStatic
    def getNumberOfFreeParameters(self):
        """
        :rtype: int
        """
        # For now, just ignore. Not important.
        return 0

    # noinspection PyShadowingBuiltins
    def forward(self, input):
        """
        :param input: tuple of input matrices of format (input_size,time). we ignore them.
        :return: single output matrix of format (output_size,time)
        """
        if Verbose:
            print(
                "RETURNN SprintControl[pid %i] SprintNnPythonLayer.forward: %s"
                % (os.getpid(), input[0].shape if input else repr(input)[:10])
            )
        assert len(input) == 1
        assert input[0].ndim == 2
        assert input[0].shape[0] == self.input_size
        seg_len = input[0].shape[1]
        posteriors = PythonControl.instance.get_current_seg_posteriors(seg_len=seg_len)  # (time,label)
        if PythonControl.instance.posteriors_in_log_space:
            assert PythonControl.instance.sprint_knows_about_log_space_probs
        assert posteriors.shape == (seg_len, self.output_size)
        return posteriors.T

    # noinspection PyPep8Naming
    def backpropagate(self, errorSignalIn):
        """
        :param numpy.ndarray errorSignalIn: matrix of format (output_size,time)
        :return: tuple of matrices of format (input_size,time)
        :rtype: numpy.ndarray
        """
        if Verbose:
            print(
                "RETURNN SprintControl[pid %i] SprintNnPythonLayer.backpropagate: %r"
                % (os.getpid(), errorSignalIn.shape)
            )
        assert errorSignalIn.ndim == 2
        assert errorSignalIn.shape[0] == self.output_size
        seg_len = errorSignalIn.shape[1]
        PythonControl.instance.set_current_seg_error_signal(seg_len=seg_len, error_signal=errorSignalIn.T)
        # must return a 1-tuple
        return (numpy.zeros((self.input_size, seg_len), dtype="float32"),)  # noqa


# End SprintNnPythonLayer. }


class PythonControl:
    """
    This will send data to RETURNN over a pipe.
    We expect that we are child process and the parent process has spawned us,

    An instance of this class is also the interface for multiple Sprint interfaces, i.e.:
      * PythonControl (standalone via NnTrainer tool)
      * PythonControl (via SegmentwiseNnTrainer)
      * implicitly PythonSegmentOrder (see code above)
    """

    Version = 1  # increase when some protocol changes
    instance = None  # type: typing.Optional[PythonControl]

    @classmethod
    def create(cls, **kwargs):
        """
        :param kwargs: passed to :class:`PythonControl`
        :rtype: PythonControl
        """
        if cls.instance:
            cls.instance._additional_init(**kwargs)
            return cls.instance
        print("RETURNN SprintControl[pid %i] PythonControl create %r" % (os.getpid(), kwargs))
        return PythonControl(**kwargs)

    def __init__(self, c2p_fd, p2c_fd, **kwargs):
        """
        :param int c2p_fd: child-to-parent file descriptor
        :param int p2c_fd: parent-to-child file descriptor
        """
        print("RETURNN SprintControl[pid %i] PythonControl init %r" % (os.getpid(), kwargs))
        assert not self.__class__.instance, "only one instance expected"
        self.__class__.instance = self
        self.cond = Condition()
        self.pipe_c2p = os.fdopen(c2p_fd, "wb")
        self.pipe_p2c = os.fdopen(p2c_fd, "rb")
        self.sprint_callback = None  # via self._init
        self.sprint_version_number = None  # via self._init
        self.callback = None  # either via Sprint, or self.own_threaded_callback
        self.loss_and_error_signal_via_sprint_callback = False
        # So, we get posteriors here from SprintErrorSignals. This will give us always log-probs for now.
        self.posteriors_in_log_space = True  # right now, not configurable
        # We directly forward these posteriors like they are to Sprint. So we assume
        # that Sprint knows about this. For CTC, this is e.g. the option --*.input-in-log-space=true.
        # The Sequence training setup would have a special BiasLayer after the PythonLayer
        # and Sprint assumes log-probs in that case.
        # For other cases, Sprint might not know. We can easily handle that here.
        self.sprint_knows_about_log_space_probs = True  # right now, not configurable
        self.control_loop_started = False
        self.control_loop_exited = False
        self.control_thread__have_new_seg = False
        self.control_thread__have_new_error_signal = False
        self.seg_name = None
        self.seg_len = None
        self.posteriors = None
        self.asked_for_posteriors = False
        self.notified_for_segment = False
        self.error_signal = None
        self.loss = None
        self._init(**kwargs)

    def _additional_init(self, **kwargs):
        print("RETURNN SprintControl[pid %i] PythonControl additional_init %r" % (os.getpid(), kwargs))
        self._init(**kwargs)

    # noinspection PyUnusedLocal
    def _init(self, name, sprint_unit=None, callback=None, version_number=None, min_version_number=None, **kwargs):
        if name == "Sprint.PythonControl":
            print("RETURNN SprintControl[pid %i] init for Sprint.PythonControl %r" % (os.getpid(), kwargs))
            assert min_version_number
            assert (version_number or 0) >= min_version_number, "need new Sprint"
            self.sprint_version_number = version_number
            if callback:
                self.sprint_callback = callback

    # noinspection PyUnusedLocal
    def init_processing(self, input_dim=None, output_dim=None, **kwargs):
        """
        This is called via Sprint when we use PythonControl to iterate the corpus,
        i.e. we set --*.action=python-control in Sprint in the NN trainer tool.
        We expect that we use the Sprint callback to calculate loss and error signal.
        This is called on the first segment.
        input_dim/output_dim are set iff we extract features/alignments.

        :param int|None input_dim:
        :param int|None output_dim:
        """
        print(
            "RETURNN SprintControl[pid %i] init_processing input_dim=%r, output_dim=%r"
            % (os.getpid(), input_dim, output_dim)
        )
        print("RETURNN SprintControl[pid %i] loss_and_error_signal_via_sprint_callback enabled" % (os.getpid(),))
        self.loss_and_error_signal_via_sprint_callback = True
        assert self.sprint_callback

    # noinspection PyUnusedLocal
    def process_segment(self, name, orthography, features=None, alignment=None, soft_alignment=None, **kwargs):
        """
        This is called via Sprint when we use PythonControl to iterate the corpus.

        :param str name: segment name
        :param str orthography: segment orth
        :param numpy.ndarray|None features:
        :param numpy.ndarray|None alignment:
        :param numpy.ndarray|None soft_alignment:
        """
        if Verbose:
            print(
                "RETURNN SprintControl[pid %i] process_segment name=%r orth=%r"
                % (os.getpid(), name, orthography[:10] + "...")
            )
        assert self.loss_and_error_signal_via_sprint_callback
        assert self.seg_name == name  # via self.handle_cmd_get_loss_and_error_signal()
        assert self.posteriors.ndim == 2  # (time,dim)
        assert features is None, "in Sprint, set --*.extract-features=false"
        assert alignment is None, "in Sprint, set --*.extract-alignment=false"
        assert soft_alignment is None, "in Sprint, set --*.extract-alignment=false"
        loss, error_signal = self._get_loss_and_error_signal_via_sprint_callback(
            seg_name=name, orthography=orthography, posteriors=self.posteriors
        )
        assert loss is not None
        assert error_signal is not None
        with self.cond:
            self.loss = loss
            self.error_signal = error_signal
            self.cond.notify_all()

    def _get_loss_and_error_signal_via_sprint_callback(self, seg_name, orthography, posteriors):
        """
        :param str seg_name:
        :param str orthography:
        :param numpy.ndarray posteriors:
        :return: (loss, error_signal)
        :rtype: (float, numpy.ndarray)
        """
        if self.posteriors_in_log_space:
            if self.sprint_knows_about_log_space_probs:  # --*.input-in-log-space=true for CTC in Sprint
                # We assume that Sprint will return gradient for activation before softmax, e.g. y - \hat{y} for CE/CTC.
                output_error_type = "error-signal"
            else:
                posteriors = numpy.exp(posteriors)
                output_error_type = "error-signal-before-softmax"
        else:
            output_error_type = "error-signal-before-softmax"
        loss, error_signal = self.sprint_callback(
            "calculate_criterion",
            posteriors=posteriors.T,
            orthography=orthography,  # Sprint wants shape (dim,time).
            output_error_type=output_error_type,
            segment_name=seg_name,
        )
        if loss is None:
            return self._default_skipped_loss(), self._default_skipped_error_signal(posteriors)
        error_signal = error_signal.T  # Sprint returns (dim,time) shape.
        assert error_signal.shape == posteriors.shape
        return loss, error_signal

    def _send(self, data):
        util.write_pickled_object(self.pipe_c2p, data)

    def _read(self):
        return util.read_pickled_object(self.pipe_p2c)

    def close(self):
        """
        Close pipe.
        """
        self.pipe_c2p.close()
        self.pipe_p2c.close()

    def _handle_cmd_exit(self):
        self.close()
        raise SystemExit

    # noinspection PyUnusedLocal
    def _handle_cmd_init(self, name, version):
        assert version == self.Version
        return "SprintControl", self.Version

    def _handle_cmd_get_loss_and_error_signal(self, seg_name, seg_len, posteriors):
        """
        :param str seg_name: seg name
        :param int seg_len: the segment length in frames
        :param numpy.ndarray posteriors: 2d (time,label) float array

        See SprintErrorSignals.SprintSubprocessInstance.get_loss_and_error_signal().
        """
        assert isinstance(seg_len, (int, long, numpy.int32))
        assert seg_len > 0
        assert posteriors.ndim == 2
        assert posteriors.shape[0] == seg_len
        if Verbose:
            print(
                "RETURNN SprintControl[pid %i] PythonControl handle_cmd_get_loss_and_error_signal: name=%r, len=%r"
                % (os.getpid(), seg_name, seg_len)
            )
        with self.cond:
            self.control_thread__have_new_seg = True
            self.control_thread__have_new_error_signal = False
            if isinstance(seg_name, bytes):
                self.seg_name = seg_name.decode("utf-8")
            else:
                self.seg_name = seg_name
            self.seg_len = seg_len
            self.posteriors = posteriors
            self.error_signal = None
            self.loss = None
            self.asked_for_posteriors = False
            self.notified_for_segment = False
            self.cond.notify_all()
        loss, error_signal = self.callback("get_loss_and_error_signal", seg_name, seg_len, posteriors)
        assert error_signal.shape == posteriors.shape
        with self.cond:
            self.control_thread__have_new_error_signal = True
            self.posteriors = None
            self.cond.notify_all()
        numpy_set_unused(posteriors)
        error_signal = error_signal.astype("float32", copy=False)
        return loss, error_signal

    def _handle_cmd_export_allophone_state_fsa_by_segment_name(self, segment_name):
        return self.callback("export_allophone_state_fsa_by_segment_name", segment_name)

    def _handle_cmd(self, cmd, *args):
        """
        :param str cmd:
        :param args:
        :return: some tuple, whatever the func returns
        :rtype: tuple
        """
        func = getattr(self, "_handle_cmd_%s" % cmd)
        return func(*args)

    def handle_next(self):
        """
        Called by self.run_control_loop.
        We catch some message from our parent process, handle it and send back the result.
        """
        import sys

        args = self._read()
        try:
            if not isinstance(args, tuple):
                raise TypeError("expected tuple but got %r" % args)
            if len(args) < 1:
                raise Exception("need multiple args (cmd, ...)")
            res = self._handle_cmd(*args)
        except Exception as e:
            print("RETURNN SprintControl[pid %i] PythonControl handle_next exception" % (os.getpid(),))
            sys.excepthook(*sys.exc_info())
            self._send(("exception", str(e)))
        else:
            assert isinstance(res, tuple)
            self._send(("ok",) + res)

    def run_control_loop(self, callback, **kwargs):
        """
        Called by Sprint when we are in PythonControl run_control_loop mode.
        Also called by us via self.run_threaded_control_loop().
        """
        print("RETURNN SprintControl[pid %i] PythonControl run_control_loop: %r, %r" % (os.getpid(), callback, kwargs))
        print(
            "RETURNN SprintControl[pid %i] PythonControl run_control_loop control: %r"
            % (os.getpid(), callback("version"))
        )
        self.callback = callback
        with self.cond:
            assert not self.control_loop_started
            self.control_loop_started = True
            self.cond.notify_all()
        try:
            while True:
                self.handle_next()
        finally:
            with self.cond:
                self.control_loop_exited = True
                self.cond.notify_all()

    # noinspection PyMethodMayBeStatic
    def exit(self, **kwargs):
        """
        Called by Sprint.
        """
        print("RETURNN SprintControl[pid %i] PythonControl exit: %r" % (os.getpid(), kwargs))

    def check_control_loop_running(self):
        """
        Called by Sprint.
        """
        if self.control_loop_started:
            print(
                "RETURNN SprintControl[pid %i] PythonControl check_control_loop_running: already running"
                % (os.getpid(),)
            )
            return
        self.run_threaded_control_loop()

    def run_threaded_control_loop(self):
        """
        Called by Sprint.
        """
        print("RETURNN SprintControl[pid %i] PythonControl run_threaded_control_loop" % (os.getpid(),))
        from threading import Thread

        def control_loop():
            """
            Control loop.
            """
            rnn.init_better_exchook()
            self.run_control_loop(self.own_threaded_callback)

        t = Thread(target=control_loop, name="SprintControl.PythonControl.threaded_control_loop")
        t.daemon = True
        t.start()
        while True:
            with self.cond:
                if self.control_loop_started:
                    return
                assert t.is_alive()
                self.cond.wait(timeout=1)

    def own_threaded_callback(self, cmd, *args):
        """
        This is used if we run our own control loop via run_threaded_control_loop.
        """
        func = getattr(self, "own_tcb_%s" % cmd)
        return func(*args)

    # noinspection PyMethodMayBeStatic
    def own_tcb_version(self):
        """
        :return: version string
        :rtype: str
        """
        return "<version>RETURNN.own_threaded_callback</version>"

    # noinspection PyUnusedLocal
    def own_tcb_get_loss_and_error_signal(self, seg_name, seg_len, posteriors):
        """
        :param seg_name:
        :param seg_len:
        :param posteriors:
        :return:
        """
        # Wait until we get the loss and error signal.
        while True:
            with self.cond:
                if self.loss is not None and self.error_signal is not None:
                    return self.loss, self.error_signal
                self.cond.wait(timeout=1)

    def init_segment(self, segment_name):
        """
        Called by Sprint PythonControl in FeedForwardTrainer/SegmentwiseNnTrainer.
        """
        if Verbose:
            print("RETURNN SprintControl[pid %i] init_segment %s" % (os.getpid(), segment_name))
        with self.cond:
            assert self.seg_name == segment_name
            self.notified_for_segment = True
            self.cond.notify_all()

    def notify_segment_loss(self, segment_name, loss):
        """
        Called by Sprint PythonControl in FeedForwardTrainer/SegmentwiseNnTrainer.
        """
        if Verbose:
            print("RETURNN SprintControl[pid %i] notify_segment_loss %s %s" % (os.getpid(), segment_name, loss))
        self.set_current_seg_loss(seg_name=segment_name, loss=loss)

    def get_current_seg_posteriors(self, seg_len):
        """
        :param int seg_len: just for double checking, the length of the current segment
        :return: matrix (time,label)
        """
        with self.cond:
            assert self.seg_len == seg_len
            assert self.posteriors.shape[0] == seg_len
            self.asked_for_posteriors = True
            self.cond.notify_all()
            return self.posteriors

    def set_current_seg_error_signal(self, seg_len, error_signal):
        """
        :param int seg_len: just for double checking, the length of the current segment
        :param error_signal: matrix (time,label)
        """
        with self.cond:
            assert self.seg_len == seg_len
            assert error_signal.ndim == 2
            assert error_signal.shape[0] == seg_len
            self.error_signal = error_signal
            self.cond.notify_all()

    def set_current_seg_loss(self, seg_name, loss):
        """
        :param str|None seg_name: just for double checking, the name of the current segment. might be None
        :param float loss: the loss of the current seg
        """
        with self.cond:
            if seg_name:
                assert self.seg_name == seg_name
            self.loss = loss
            self.cond.notify_all()

    # noinspection PyMethodMayBeStatic
    def _default_skipped_loss(self):
        """
        :rtype: float
        """
        return float("inf")

    # noinspection PyMethodMayBeStatic
    def _default_skipped_error_signal(self, posteriors):
        """
        :param numpy.ndarray posteriors:
        :rtype: numpy.ndarray
        """
        return numpy.zeros_like(posteriors)

    def _skip_segment_loss_and_error(self):
        with self.cond:
            assert self.posteriors is not None
            if self.loss is None:
                self.loss = self._default_skipped_loss()
            if self.error_signal is None:
                self.error_signal = self._default_skipped_error_signal(self.posteriors)
            self.cond.notify_all()

    def _wait_for_control_loop_error_signal(self):
        while True:
            with self.cond:
                if self.control_thread__have_new_error_signal or self.control_thread__have_new_seg:
                    break
                if self.control_loop_exited:
                    break
                if self.loss is None or self.error_signal is None:
                    break
                if Verbose:
                    print(
                        "RETURNN SprintControl[pid %i] getSegmentList: wait for control loop to handle error signal"
                        % (os.getpid(),)
                    )
                self.cond.wait(timeout=1)

    def segment_list_iterator(self):
        """
        :return: yields segment names
        :rtype: typing.Iterator[str]
        """
        with self.cond:
            assert self.control_loop_started

        while True:  # outer loop
            # wait until we get new segment
            while True:
                with self.cond:
                    if self.control_thread__have_new_seg:
                        assert self.seg_name
                        seg_name = self.seg_name
                        self.control_thread__have_new_seg = False
                        break
                    if self.control_loop_exited:
                        return  # no more segments
                    self.cond.wait(timeout=1)

            # We got a new segment name from the parent RETURNN process
            # (via self.handle_cmd_get_loss_and_error_signal()).
            # We wait in this segment because we wait to get the error signal from Sprint
            # (via SprintNnPythonLayer.backpropagate()).
            # Sprint waits currently for us to get the new segment (in the PythonSegmentOrder code).
            # Once it gets it, it will call SprintNnPythonLayer.forward(), then calculate the loss and error signal
            # and then call SprintNnPythonLayer.backpropagate().
            if Verbose:
                print("RETURNN SprintControl[pid %i] getSegmentList, yield %r" % (os.getpid(), seg_name))
            yield seg_name

            # We might need to wait for the control loop thread.
            self._wait_for_control_loop_error_signal()

            # When we are back here, Sprint asks for the next segment.
            # It means that is has finished any processing with this segment.
            with self.cond:
                # See self.handle_cmd_get_loss_and_error_signal().
                # We are still stuck in there in the other thread, if not self.have_new_error_signal.
                # Maybe the PythonLayer was not used?
                # Or Sprint could not calculate the criterion for this segment (bad lattice or so).
                if not (self.control_thread__have_new_error_signal or self.control_thread__have_new_seg):
                    print(
                        "RETURNN SprintControl[pid %i] getSegmentList, no error signal, skip segment: %s"
                        % (os.getpid(), seg_name)
                    )
                    if Verbose:
                        # Print Sprint stacktrace.
                        import signal

                        os.kill(os.getpid(), signal.SIGUSR1)
                    if not self.notified_for_segment:
                        print(
                            (
                                "RETURNN SprintControl[pid %i] getSegmentList: "
                                "Do you use PythonControl in the Sprint trainer? Got no segment notification."
                            )
                            % (os.getpid(),)
                        )
                    if not self.asked_for_posteriors:
                        print(
                            (
                                "RETURNN SprintControl[pid %i] getSegmentList: "
                                "Do you use PythonLayer in Sprint? Did not get asked for posteriors."
                            )
                            % (os.getpid(),)
                        )
                    self._skip_segment_loss_and_error()
                    self._wait_for_control_loop_error_signal()



"""
Import the parameters from the HuggingFace Llama model (PyTorch).
"""

from __future__ import annotations
from typing import TYPE_CHECKING, Union
import returnn.frontend as rf
from returnn.frontend.decoder.transformer import TransformerDecoder, TransformerDecoderLayer, FeedForwardGated

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences,PyPackageRequirements,PyProtectedMember
    from transformers.models.llama.modeling_llama import (
        LlamaModel,
        LlamaForCausalLM,
        LlamaDecoderLayer,
        LlamaMLP,
        LlamaRMSNorm,
        LlamaAttention,
    )


def import_params_hf_llama_to_rf_transformer_decoder(
    model_hf: Union[LlamaModel, LlamaForCausalLM], model_rf: TransformerDecoder
):
    """
    Import params from HF Llama model to RF :class:`TransformerDecoder`.
    """
    import torch

    # noinspection PyUnresolvedReferences,PyPackageRequirements,PyProtectedMember
    from transformers.models.llama.modeling_llama import LlamaModel, LlamaForCausalLM, LlamaDecoderLayer

    print("HF Model:")
    print(model_hf)
    print("Parameters:")
    num_params_hf = 0
    for k, v in model_hf.named_parameters():
        print(f"{k}: {list(v.shape)} {v.dtype}")
        num_params_hf += v.numel()
    print("Total number of parameters:", num_params_hf)

    print("RF Model:")
    print(model_rf)
    print("Parameters:")
    num_params_rf = 0
    for k, v in model_rf.named_parameters():
        print(f"{k}: {list(v.dims)} {v.dtype}")
        assert isinstance(v.raw_tensor, torch.nn.Parameter)
        num_params_rf += v.num_elements()
    print("Total number of parameters:", num_params_rf)
    # Check if the number of parameters is the same below.
    # First import individual sub modules.
    # We might detect any mismatches there, and this will easy the debugging.

    lm_head = None
    if isinstance(model_hf, LlamaForCausalLM):
        lm_head = model_hf.lm_head
        model_hf = model_hf.model
    else:
        # Exclude logits.
        num_params_rf -= model_rf.logits.weight.num_elements()
    assert isinstance(model_hf, LlamaModel)
    assert model_hf.norm.weight.shape[0] == model_rf.model_dim.dimension

    assert len(model_hf.layers) == len(model_rf.layers)
    for i, (layer_hf, layer_rf) in enumerate(zip(model_hf.layers, model_rf.layers)):
        assert isinstance(layer_hf, LlamaDecoderLayer)
        assert isinstance(layer_rf, TransformerDecoderLayer)
        import_params_hf_llama_decoder_layer_to_rf(layer_hf, layer_rf)

    assert model_hf.embed_tokens.weight.shape == model_rf.input_embedding.weight.raw_tensor.shape
    with torch.no_grad():
        model_rf.input_embedding.weight.raw_tensor.copy_(model_hf.embed_tokens.weight)  # (vocab,hidden)

    assert isinstance(model_rf.final_layer_norm, rf.RMSNorm)
    import_params_hf_llama_rms_norm_to_rf(model_hf.norm, model_rf.final_layer_norm)

    if lm_head is not None:
        assert lm_head.bias is None and model_rf.logits.bias is None  # not implemented
        # Torch Linear: (out,in), but RF has (in,out).
        with torch.no_grad():
            model_rf.logits.weight.raw_tensor.copy_(lm_head.weight.T)  # (hidden,vocab)

    assert num_params_rf == num_params_hf, f"missmatch num params: RF {num_params_rf} != HF {num_params_hf}"


def import_params_hf_llama_decoder_layer_to_rf(model_hf: LlamaDecoderLayer, model_rf: TransformerDecoderLayer):
    """
    Import the parameters from the HF Llama decoder layer.
    """
    import torch

    assert model_hf.hidden_size == model_rf.out_dim.dimension

    print("HF Model:")
    print(model_hf)
    print("Parameters:")
    num_params_hf = 0
    for k, v in model_hf.named_parameters():
        print(f"{k}: {list(v.shape)} {v.dtype}")
        num_params_hf += v.numel()
    print("Total number of parameters:", num_params_hf)

    print("RF Model:")
    print(model_rf)
    print("Parameters:")
    num_params_rf = 0
    for k, v in model_rf.named_parameters():
        print(f"{k}: {list(v.dims)} {v.dtype}")
        assert isinstance(v.raw_tensor, torch.nn.Parameter)
        num_params_rf += v.num_elements()
    print("Total number of parameters:", num_params_rf)
    # Check if the number of parameters is the same below.
    # First import individual sub modules.
    # We might detect any mismatches there, and this will easy the debugging.

    assert isinstance(model_rf.ff, FeedForwardGated), f"unexpected: {model_rf.ff}"
    import_params_hf_llama_mlp_to_rf_feed_forward_gated(model_hf.mlp, model_rf.ff)

    assert isinstance(model_rf.self_att, rf.RotaryPosCausalSelfAttention), f"unexpected: {model_rf.self_att}"
    import_params_hf_llama_att_to_rf_rotary_att(model_hf.self_attn, model_rf.self_att)

    assert isinstance(model_rf.self_att_layer_norm, rf.RMSNorm), f"unexpected: {model_rf.self_att_layer_norm}"
    import_params_hf_llama_rms_norm_to_rf(model_hf.input_layernorm, model_rf.self_att_layer_norm)

    assert isinstance(model_rf.ff_layer_norm, rf.RMSNorm), f"unexpected: {model_rf.ff_layer_norm}"
    import_params_hf_llama_rms_norm_to_rf(model_hf.post_attention_layernorm, model_rf.ff_layer_norm)

    assert num_params_rf == num_params_hf


def import_params_hf_llama_mlp_to_rf_feed_forward_gated(model_hf: LlamaMLP, model_rf: FeedForwardGated):
    """
    Import the parameters from the HF Llama MLP module.
    """
    import torch

    assert model_hf.hidden_size == model_rf.out_dim.dimension == model_rf.linear_ff.in_dim.dimension

    print("HF Model:")
    print(model_hf)
    print("Parameters:")
    num_params_hf = 0
    for k, v in model_hf.named_parameters():
        print(f"{k}: {list(v.shape)} {v.dtype}")
        num_params_hf += v.numel()
    print("Total number of parameters:", num_params_hf)

    print("RF Model:")
    print(model_rf)
    print("Parameters:")
    num_params_rf = 0
    for k, v in model_rf.named_parameters():
        print(f"{k}: {list(v.dims)} {v.dtype}")
        assert isinstance(v.raw_tensor, torch.nn.Parameter)
        num_params_rf += v.num_elements()
    print("Total number of parameters:", num_params_rf)
    assert num_params_rf == num_params_hf

    # Torch Linear: (out,in), but RF has (in,out).
    w1 = model_hf.gate_proj.weight.T  # (in,out)
    w2 = model_hf.up_proj.weight.T  # (in,out)
    w3 = model_hf.down_proj.weight.T  # (out,in)
    assert model_hf.gate_proj.bias is None  # not implemented
    assert model_hf.up_proj.bias is None  # not implemented
    assert model_hf.down_proj.bias is None  # not implemented
    with torch.no_grad():
        w = torch.cat((w1, w2), dim=1)  # (in,out*2)
        model_rf.linear_ff.weight.raw_tensor.copy_(w)
        model_rf.linear_out.weight.raw_tensor.copy_(w3)


def import_params_hf_llama_rms_norm_to_rf(model_hf: LlamaRMSNorm, model_rf: rf.RMSNorm):
    """
    Import the parameters from the HF Llama RMSNorm module.
    """
    import torch

    assert model_hf.weight.shape[0] == model_rf.in_dim.dimension

    print("HF Model:")
    print(model_hf)
    print("Parameters:")
    num_params_hf = 0
    for k, v in model_hf.named_parameters():
        print(f"{k}: {list(v.shape)} {v.dtype}")
        num_params_hf += v.numel()
    print("Total number of parameters:", num_params_hf)

    print("RF Model:")
    print(model_rf)
    print("Parameters:")
    num_params_rf = 0
    for k, v in model_rf.named_parameters():
        print(f"{k}: {list(v.dims)} {v.dtype}")
        assert isinstance(v.raw_tensor, torch.nn.Parameter)
        num_params_rf += v.num_elements()
    print("Total number of parameters:", num_params_rf)
    assert num_params_rf == num_params_hf

    w = model_hf.weight  # (in,)
    with torch.no_grad():
        model_rf.scale.raw_tensor.copy_(w)


def import_params_hf_llama_att_to_rf_rotary_att(model_hf: LlamaAttention, model_rf: rf.RotaryPosCausalSelfAttention):
    """
    Import the parameters from the HF Llama attention module.
    """
    import torch

    assert model_hf.config.num_attention_heads == model_rf.num_heads.dimension
    assert model_hf.config.hidden_size == model_rf.in_dim.dimension
    dim = model_hf.config.hidden_size
    nh = model_hf.config.num_attention_heads
    hdim = dim // nh

    print("HF Model:")
    print(model_hf)
    print("Parameters:")
    num_params_hf = 0
    for k, v in model_hf.named_parameters():
        print(f"{k}: {list(v.shape)} {v.dtype}")
        num_params_hf += v.numel()
    print("Total number of parameters:", num_params_hf)

    print("RF Model:")
    print(model_rf)
    print("Parameters:")
    num_params_rf = 0
    for k, v in model_rf.named_parameters():
        print(f"{k}: {list(v.dims)} {v.dtype}")
        assert isinstance(v.raw_tensor, torch.nn.Parameter)
        num_params_rf += v.num_elements()
    print("Total number of parameters:", num_params_rf)
    assert num_params_rf == num_params_hf, f"num params RF {num_params_rf} != params HF {num_params_hf}"

    # Torch Linear: (out,in), but RF has (in,out).
    q = model_hf.q_proj.weight.T.reshape(dim, nh, hdim)  # (in,h,out/h)
    k = model_hf.k_proj.weight.T.reshape(dim, nh, hdim)  # (in,h,out/h)
    v = model_hf.v_proj.weight.T.reshape(dim, nh, hdim)  # (in,h,out/h)
    q = q.reshape(dim, nh, 2, hdim // 2).transpose(-1, -2).flatten(-2)  # reorder complex numbers
    k = k.reshape(dim, nh, 2, hdim // 2).transpose(-1, -2).flatten(-2)  # reorder complex numbers
    qkv = torch.cat([q, k, v], dim=2)  # (in,h,out/h*3)
    qkv = qkv.reshape(dim, 3 * dim)
    assert model_hf.q_proj.bias is None  # not implemented
    with torch.no_grad():
        model_rf.qkv.weight.raw_tensor.copy_(qkv)
        model_rf.proj.weight.raw_tensor.copy_(model_hf.o_proj.weight.T)



"""
Used by setup.py.
"""

from __future__ import annotations
from pprint import pprint
import os
import sys


_my_dir = os.path.dirname(os.path.abspath(__file__))
# Use realpath to resolve any symlinks. We want the real root-dir, to be able to check the Git revision.
_root_dir = os.path.dirname(os.path.realpath(_my_dir))


def debug_print_file(fn):
    """
    :param str fn:
    """
    print("%s:" % fn)
    if not os.path.exists(fn):
        print("<does not exist>")
        return
    if os.path.isdir(fn):
        print("<dir:>")
        pprint(sorted(os.listdir(fn)))
        return
    print(open(fn).read())


def parse_pkg_info(fn):
    """
    :param str fn:
    :return: dict with info written by distutils. e.g. ``res["Version"]`` is the version.
    :rtype: dict[str,str]
    """
    res = {}
    for ln in open(fn).read().splitlines():
        if not ln or not ln[:1].strip():
            continue
        key, value = ln.split(": ", 1)
        res[key] = value
    return res


def git_head_version(git_dir=_root_dir, long=False):
    """
    :param str git_dir:
    :param bool long: see :func:`get_version_str`
    :rtype: str
    """
    from returnn.util.basic import git_commit_date, git_commit_rev, git_is_dirty

    commit_date = git_commit_date(git_dir=git_dir)  # like "20190202.154527"
    version = "1.%s" % commit_date  # distutils.version.StrictVersion compatible
    if long:
        # Keep SemVer compatible.
        rev = git_commit_rev(git_dir=git_dir)
        version += "+git.%s" % rev
        if git_is_dirty(git_dir=git_dir):
            version += ".dirty"
    return version


def get_version_str(verbose=False, verbose_error=False, fallback=None, long=False):
    """
    :param bool verbose: print exactly how we end up with some version
    :param bool verbose_error: print only any potential errors
    :param str|None fallback:
    :param bool long:
      False: Always distutils.version.StrictVersion compatible. just like "1.20190202.154527".
      True: Will also add the revision string, like "1.20180724.141845+git.7865d01".
        The format might change in the future.
        We will keep it `SemVer <https://semver.org/>`__ compatible.
        I.e. the string before the `"+"` will be the short version.
        We always make sure that there is a `"+"` in the string.
    :rtype: str
    """
    # Earlier we checked PKG-INFO, via parse_pkg_info. Both in the root-dir and in my-dir.
    # Now we should always have _setup_info_generated.py, copied by our own setup.
    # Do not use PKG-INFO at all anymore (for now), as it would only have the short version.
    # Only check _setup_info_generated in the current dir, not in the root-dir,
    # because we want to only use it if this was installed via a package.
    # Otherwise, we want the current Git version.
    if os.path.exists("%s/_setup_info_generated.py" % _my_dir):
        # noinspection PyUnresolvedReferences
        from . import _setup_info_generated as info

        if verbose:
            print("Found _setup_info_generated.py, long version %r, version %r." % (info.long_version, info.version))
        if long:
            assert "+" in info.long_version
            return info.long_version
        return info.version

    info_in_root_filename = "%s/_setup_info_generated.py" % _root_dir
    if os.path.exists(info_in_root_filename):
        # The root dir might not be in sys.path, so just load directly.
        code = compile(open(info_in_root_filename).read(), info_in_root_filename, "exec")
        info = {}
        eval(code, info)
        version = info["version"]
        long_version = info["long_version"]
        if verbose:
            print("Found %r in root, long version %r, version %r." % (info_in_root_filename, long_version, version))
        if long:
            assert "+" in long_version
            return long_version
        return version

    if os.path.exists("%s/.git" % _root_dir):
        try:
            version = git_head_version(git_dir=_root_dir, long=long)
            if verbose:
                print("Version via Git:", version)
            if long:
                assert "+" in version
            return version
        except Exception as exc:
            if verbose or verbose_error:
                print("Exception while getting Git version:", exc)
                sys.excepthook(*sys.exc_info())
            if not fallback:
                raise  # no fallback

    if fallback:
        if verbose:
            print("Version via fallback:", fallback)
        if long:
            assert "+" in fallback
        return fallback
    raise Exception("Cannot get RETURNN version.")



"""
This module covers the optimizer (SGD, Adam, etc) logic,
and model param update logic in general.
"""

from __future__ import annotations

from typing import Optional, Union, Any, Type, Callable, Sequence, Iterable, Iterator, Set, Dict, List, Tuple
import os
import gc
import torch
import typing

import returnn
from returnn.log import log
from returnn.util.basic import RefIdEq, get_fwd_compat_kwargs
import returnn.frontend as rf
from returnn.torch.frontend.bridge import wrapped_pt_module_to_rf_module

_OptimizerClassesDictInitialized = False
_OptimizerClassesDict = {}


def _init_optimizer_classes_dict():
    """
    Initializes a global dictionary with all optimizers available in PyTorch.
    """
    global _OptimizerClassesDictInitialized
    if _OptimizerClassesDictInitialized:
        return
    _OptimizerClassesDictInitialized = True
    for name, cls in list(vars(torch.optim).items()):
        assert isinstance(name, str)
        # Check if cls is a valid subclass of torch.optim.Optimizer
        if not isinstance(cls, type) or not issubclass(cls, torch.optim.Optimizer):
            continue
        assert name not in _OptimizerClassesDict
        _OptimizerClassesDict[name.lower()] = cls


def get_optimizer_class(
    class_name: Union[str, Type[torch.optim.Optimizer], Callable[[], Type[torch.optim.Optimizer]]],
) -> Type[torch.optim.Optimizer]:
    """
    :param class_name: Optimizer class, either as str (e.g. "adam"), as type (torch.optim.Adam) or callable.
        If str, we support all torch.optim optimizers (ignoring case) (e.g. "adam"),
        or class names with full module path (e.g. "returnn.torch.optim.lion.Lion").
    :return: Optimizer class, e.g. torch.optim.Adam
    """
    _init_optimizer_classes_dict()
    if isinstance(class_name, type):
        assert issubclass(class_name, torch.optim.Optimizer)
        return class_name
    elif callable(class_name):
        return class_name()
    elif isinstance(class_name, str):
        if "." in class_name:
            import importlib

            mod_name, class_name_ = class_name.rsplit(".", 1)
            mod = importlib.import_module(mod_name)
            return getattr(mod, class_name_)

        if class_name.lower() not in _OptimizerClassesDict:
            raise ValueError(
                "Optimizer %r not found in the available torch optimizers list: %s."
                % (
                    class_name.lower(),
                    ", ".join("'%s'" % key for key in _OptimizerClassesDict),
                )
            )
        return _OptimizerClassesDict[class_name.lower()]
    else:
        raise TypeError(f"Invalid optimizer class_name {class_name!r} type {type(class_name).__name__}")


def _get_class_init_kwargs(optim_class):
    """
    Obtains the keyword arguments of the class provided as parameter that the user can add to their optimizer.

    :param type[torch.optim.Optimizer] optim_class: Optimizer class.
    :return: Keyword arguments of the provided class.
    :rtype: List[str]
    """
    from returnn.util.basic import collect_class_init_kwargs

    optim_class_init_kwargs = collect_class_init_kwargs(optim_class)
    # We already provide params by default, remove it so that the user doesn't add it to the optimizer dict.
    optim_class_init_kwargs.remove("params")

    return optim_class_init_kwargs


class Updater:
    """
    Wraps a torch.optim.Optimizer, and extends it by some further functionality.
    """

    def __init__(self, *, config, network, device, initial_learning_rate=1.0):
        """
        :param returnn.config.Config config: config defining the training conditions.
        :param torch.nn.Module network: PyTorch Module defining the network.
        :param torch.device|str device:
        :param float initial_learning_rate:
        """
        self.config = config
        self.learning_rate = float(initial_learning_rate)
        self._effective_learning_rate = self.learning_rate
        self.network = network
        self._device = device
        # Just set the very first step as initial values here.
        # They will be overwritten via set_current_train_step() below.
        self._current_train_step = 0
        self._current_epoch = 1
        self._current_epoch_continuous = 0.0

        self.learning_rate_function = self.config.typed_value("dynamic_learning_rate", None)
        if self.learning_rate_function is not None:
            print("Using dynamic learning rate scheduler that updates based on global train steps", file=log.v2)
            if callable(self.learning_rate_function):
                import inspect

                signature = inspect.signature(self.learning_rate_function)
                assert any([arg.kind == inspect.Parameter.VAR_KEYWORD for arg in signature.parameters.values()]), (
                    "please specify **kwargs in dynamic_learning_rate for future compatibility"
                )
                if "network" in signature.parameters:
                    raise ValueError("Torch updater: dynamic_learning_rate network is TF specific")
            else:
                raise NotImplementedError("not implemented for not callable dynamic_learning_rate")

        self._optimizer_opts = None
        self.optimizer = None  # type: typing.Optional[torch.optim.Optimizer]

        self._grad_clip = self.config.float("gradient_clip", 0.0)
        self._grad_clip_global_norm = self.config.float("gradient_clip_global_norm", 0.0)
        self._grad_noise = self.config.float("gradient_noise", 0.0)

        # Check other options we have in TF updater, which we might support here later as well,
        # but currently do not support.
        for opt_name in [
            "gradient_clip_norm",
            "gradient_clip_avg_norm",
            "global_norm_tag",
            "gradient_clip_global_norm_tag",
            "grad_norm_to_clip_to_zero",
            "maximize_grad_norm",
            "debug_grad_summaries",
            "gradient_nan_inf_filter",
        ]:
            if self.config.float(opt_name, 0.0):
                raise NotImplementedError(f"PyTorch updater: option {opt_name} not supported currently")
        # Check for potential user mistakes.
        if self.config.float("grad_clip", 0.0):
            raise ValueError(
                "You set grad_clip in the config,"
                " but the option is called gradient_clip_global_norm (or other options)."
            )

        self._update_effective_learning_rate()

    def set_learning_rate(self, value):
        """
        Updates the learning rate of the optimizer at each (sub)epoch.

        :param float value: New learning rate.
        """
        self.learning_rate = float(value)
        self._update_effective_learning_rate()

    def get_effective_learning_rate(self) -> float:
        """
        :return: get the actual learning rate
        """
        return self._effective_learning_rate

    def _update_effective_learning_rate(self):
        self._effective_learning_rate = self.learning_rate
        if self.learning_rate_function is not None:
            lr = self.learning_rate_function(
                global_train_step=self._current_train_step,
                epoch=self._current_epoch,
                epoch_continuous=self._current_epoch_continuous,
                learning_rate=self.learning_rate,
                **get_fwd_compat_kwargs(),
            )
            self._effective_learning_rate = float(lr)
        if self.optimizer:
            for param_group in self.optimizer.param_groups:
                param_group["lr"] = self._effective_learning_rate

    def set_current_train_step(self, *, global_train_step: int, epoch: int, epoch_continuous: Optional[float] = None):
        """
        Obtains an updated learning rate for the current training step inside a (sub)epoch.

        :param global_train_step: Current global training step over the whole training process.
            In the first epoch, this starts at 0.
        :param epoch: Current epoch. (First epoch is 1 by RETURNN convention.)
        :param epoch_continuous: How much of the epoch is finished.
            In the first step of the first epoch, this starts at 0.0,
            and when the fist epoch is finished, this reaches 1.0,
            and the values in between are the fraction of the epoch that is finished.
            The second epoch (epoch=2) starts at 1.0,
            and when the second epoch is finished, this reaches 2.0, and so on.
            We usually calculate this based on ``epoch-1+(last_seq_idx+1)/num_seqs``,
            if the dataset can provide ``num_seqs``.
            Other schemes based on the step_idx might be used as well to calculate this,
            if the number of steps per epoch is known in advance.
        """
        self._current_train_step = global_train_step
        self._current_epoch = epoch
        self._current_epoch_continuous = epoch_continuous
        self._update_effective_learning_rate()

    def step(self, *, grad_scaler: Optional[torch.cuda.amp.GradScaler] = None):
        """
        Perform one step, i.e. update the parameters using the optimizer given the current calculated gradients.
        """
        if grad_scaler is not None:
            grad_scaler.unscale_(self.optimizer)

        if self._grad_noise:
            gradient_noise_(self.network.parameters(), self._grad_noise)
        if self._grad_clip:
            torch.nn.utils.clip_grad_value_(self.network.parameters(), self._grad_clip)
        if self._grad_clip_global_norm:
            torch.nn.utils.clip_grad_norm_(self.network.parameters(), self._grad_clip_global_norm)

        if grad_scaler is not None:
            grad_scaler.step(self.optimizer)
            grad_scaler.update()
        else:
            self.optimizer.step()

    def create_optimizer(self):
        """
        Creates an optimizer and stores it in self.optimizer.
        """
        optimizer_opts = self.config.typed_value("optimizer", None)
        if optimizer_opts is None:
            raise ValueError("config field 'optimizer' needs to be set explicitely for the Torch backend")
        self._optimizer_opts = optimizer_opts
        self.optimizer = self._create_optimizer(optimizer_opts)

    def load_optimizer(self, filename):
        """
        Loads a torch.optim.Optimizer from disk and stores it in self.optimizer.

        :param str filename: File from which to load the optimizer state.
        """
        print("Load optimizer %s" % filename, file=log.v4)
        optimizer_state = torch.load(filename, map_location=self._device)
        assert isinstance(optimizer_state, dict), f"optimizer_state is not a dict but {type(optimizer_state)}"
        if "optimizer" not in optimizer_state and "param_groups" in optimizer_state and "state" in optimizer_state:
            # Old format, convert to new format.
            optimizer_state = {"optimizer": optimizer_state}
        if optimizer_state.get("param_names") is not None:
            if len(self.optimizer.param_groups) != len(optimizer_state["optimizer"]["param_groups"]):
                raise ValueError(
                    "loaded state dict has a different number of parameter groups: ckpt %i vs. self %i"
                    % (len(optimizer_state["optimizer"]["param_groups"]), len(self.optimizer.param_groups))
                )
            # Check if we have the same parameters in the same order.
            self_param_names, param_id_to_name = self._get_opt_param_names()
            ckpt_param_names = optimizer_state["param_names"]
            if self_param_names != ckpt_param_names:
                self_param_names_dict = {name: i for i, name in enumerate(self_param_names)}
                self_param_names_critical_set = set()
                ckpt_param_names_dict = {name: i for i, name in enumerate(ckpt_param_names)}
                map_ckpt_param_idx_to_self_param_idx = {}
                self_params_not_in_ckpt = []
                self_params_not_in_ckpt_critical = []
                for param_name in self_param_names:
                    param = self.network.get_parameter(param_name)
                    if param.requires_grad:
                        self_param_names_critical_set.add(param_name)
                    if param_name not in ckpt_param_names_dict:
                        self_params_not_in_ckpt.append(param_name)
                        if param.requires_grad:
                            self_params_not_in_ckpt_critical.append(param_name)
                ckpt_params_not_in_self = []
                for i, param_name in enumerate(ckpt_param_names):
                    if param_name not in self_param_names_dict:
                        ckpt_params_not_in_self.append(param_name)
                    else:
                        map_ckpt_param_idx_to_self_param_idx[i] = self_param_names_dict[param_name]
                if self_params_not_in_ckpt_critical:
                    raise ValueError(
                        "load_optimizer: required params not in ckpt: %s" % ", ".join(self_params_not_in_ckpt_critical)
                    )
                if self_params_not_in_ckpt or ckpt_params_not_in_self:
                    print(
                        "load_optimizer: params not in ckpt: %s\n    ckpt params not existing: %s"
                        % (
                            ", ".join(self_params_not_in_ckpt) or "(None)",
                            ", ".join(ckpt_params_not_in_self) or "(None)",
                        ),
                        file=log.v3,
                    )
                    if self_params_not_in_ckpt:
                        print(
                            "load_optimizer: All params not in ckpt have required_grad=False, thus not critical.",
                            file=log.v3,
                        )
                else:
                    print("load_optimizer: Params in different order.", file=log.v3)
                print("load_optimizer: Will remap the state dict.", file=log.v3)
                for ckpt_group, self_group in zip(
                    optimizer_state["optimizer"]["param_groups"], self.optimizer.param_groups
                ):
                    # Check whether it is matching for the critical params.
                    self_group_param_names = set(param_id_to_name[id(p)] for p in self_group["params"])
                    ckpt_group_param_names = set(ckpt_param_names[p] for p in ckpt_group["params"])
                    self_group_param_names.intersection_update(self_param_names_critical_set)
                    ckpt_group_param_names.intersection_update(self_param_names_critical_set)
                    if ckpt_group_param_names != self_group_param_names:
                        raise ValueError(
                            "load_optimizer: params in group not in ckpt: %s\n  ckpt params not existing: %s"
                            % (
                                ", ".join(ckpt_group_param_names - self_group_param_names) or "(None)",
                                ", ".join(self_group_param_names - ckpt_group_param_names) or "(None)",
                            )
                        )
                    ckpt_group["params"] = [
                        self_param_names_dict[param_id_to_name[id(p)]] for p in self_group["params"]
                    ]
                optimizer_state["optimizer"]["state"] = {
                    map_ckpt_param_idx_to_self_param_idx[i]: s
                    for (i, s) in optimizer_state["optimizer"]["state"].items()
                    if i in map_ckpt_param_idx_to_self_param_idx
                }
        self.optimizer.load_state_dict(optimizer_state["optimizer"])
        # https://github.com/rwth-i6/returnn/issues/1345
        del optimizer_state
        gc.collect()

    def _get_opt_param_names(self) -> Tuple[List[str], Dict[int, str]]:
        param_id_to_name = {}  # id -> name
        for name, p in self.network.named_parameters():
            param_id_to_name[id(p)] = name
        param_names = []  # param_idx -> name
        for group in self.optimizer.param_groups:
            for p in group["params"]:
                param_names.append(param_id_to_name[id(p)])
        return param_names, param_id_to_name

    def save_optimizer(self, filename):
        """
        Saves the state of self.optimizer to a file.

        :param str filename: File in which to save the optimizer state.
        """
        directory = os.path.dirname(filename)
        if directory and not os.path.exists(directory):
            os.makedirs(directory, exist_ok=True)

        # We use optimizer.state_dict() below.
        # That will only save param order indices
        # but not the name of the parameters.
        # We also save a mapping of parameter indices to names.
        param_names, _ = self._get_opt_param_names()

        print("Save optimizer under %s" % filename, file=log.v4)
        # First write to a temp-file, to be sure that writing happens without errors,
        # and only afterward rename to the target file.
        tmp_filename = filename + ".tmp_write"
        if os.path.exists(tmp_filename):
            os.unlink(tmp_filename)
        torch.save(
            {
                "optimizer": self.optimizer.state_dict(),
                "optimizer_class_name": self.optimizer.__class__.__name__,
                "optimizer_opts": self._optimizer_opts,
                "param_names": param_names,
                "epoch": self._current_epoch,
                "step": self._current_train_step,
                "effective_learning_rate": self.get_effective_learning_rate(),
                "returnn_version": returnn.__long_version__,
            },
            tmp_filename,
        )
        os.rename(tmp_filename, filename)

    def get_optimizer(self):
        """
        :return: Wrapped optimizer object.
        :rtype: torch.optim.Optimizer
        """
        return self.optimizer

    def _create_optimizer(self, optimizer_opts):
        """
        Returns a valid optimizer considering the dictionary given by the user in the config.

        :param dict[str]|str optimizer_opts: Optimizer configuration specified by the user.
            If it's a dict, it must contain "class" with the optimizer name or callable.
            If it's a str, it must be the optimizer name.
        :return: A valid optimizer.
        :rtype: torch.optim.Optimizer
        """
        lr = self.learning_rate

        # If the parameter is already a valid optimizer, return it without further processing
        if isinstance(optimizer_opts, torch.optim.Optimizer):
            return optimizer_opts
        elif callable(optimizer_opts):
            optimizer_opts: Dict[str, Any] = {"class": optimizer_opts}
        else:
            if not isinstance(optimizer_opts, dict):
                raise ValueError("'optimizer' must of type dict, callable or torch.optim.Optimizer instance.")
            if "class" not in optimizer_opts:
                raise ValueError("'class' field of 'optimizer' dict was not set (use e.g. 'SGD', 'Adam', ...)")
            optimizer_opts = optimizer_opts.copy()

        # Resolve the optimizer class
        optim_class_name = optimizer_opts.pop("class")
        optim_class = get_optimizer_class(optim_class_name)

        # Resolve the optimizer arguments
        opt_kwargs = optimizer_opts.copy()
        optim_class_init_kwargs = _get_class_init_kwargs(optim_class)
        # epsilon is named eps in torch.
        # If the user specified it as epsilon, parse it as eps for the optimizer
        if "eps" in optim_class_init_kwargs and "epsilon" in opt_kwargs:
            opt_kwargs["eps"] = opt_kwargs.pop("epsilon")
        if "learning_rate" in opt_kwargs or "lr" in opt_kwargs:
            raise ValueError("'learning_rate' should be set outside of the 'optimizer' dict.")
        lr = lr * opt_kwargs.pop("learning_rate_multiplier", 1.0)
        opt_kwargs["lr"] = lr

        params_or_param_groups = self._get_optimizer_param_groups(optim_class, opt_kwargs)
        optimizer = optim_class(params_or_param_groups, **opt_kwargs)
        print("Optimizer: %s" % optimizer, file=log.v1)
        assert isinstance(optimizer, torch.optim.Optimizer)

        return optimizer

    def _create_default_optimizer(self):
        """
        :return: SGD optimizer.
        :rtype: torch.optim.SGD
        """
        print("Create SGD optimizer (default).", file=log.v2)
        optimizer = torch.optim.SGD(self.network.parameters(), lr=self.learning_rate)

        return optimizer

    def _get_optimizer_param_groups(
        self, optim_class: Type[torch.optim.Optimizer], optimizer_opts: Dict[str, Any]
    ) -> Union[List[Dict[str, Any]], Iterator[torch.nn.Parameter]]:
        """
        The weight_decay parameter from AdamW affects the weights of layers such as LayerNorm and Embedding.
        This function creates a blacklist of network modules and splits the optimizer groups in two:
        those who will receive weight decay, and those who won't receive it.
        The weight_decay parameter of the rest of the optimizers is L2 regularization.

        For further reading, see https://github.com/karpathy/minGPT/pull/24#issuecomment-679316025 and
        https://discuss.pytorch.org/t/weight-decay-in-the-optimizers-is-a-bad-idea-especially-with-batchnorm/16994.

        This code is based on https://github.com/karpathy/minGPT (MIT license):
        https://github.com/karpathy/minGPT/blob/3ed14b2cec0dfdad3f4b2831f2b4a86d11aef150/mingpt/model.py#L136.

        Three variants how this can be configured by the user in the optimizer options dict:

        - ``param_groups_custom``: callable which returns a list of param groups.
          This is the most flexible option, and could also go beyond just weight decay logic,
          or having more than two param groups (weight decay disabled/enabled).
        - ``weight_decay_custom_include_check``: callable which returns True/False for each param,
          to either include it in the weight decay group or not,
          or None to use the default logic.
        - ``weight_decay_modules_blacklist``: list of modules types which should not get weight decay.
          Those can be RF modules or pure PyTorch modules.
          The types can be specified as string (e.g. ``"torch.nn.LayerNorm"``) or as the type itself.

        :param optim_class: Optimizer class.
        :param optimizer_opts: Optimizer configuration specified by the user. Might be modified inplace here.
        :return: List of configurations for the different sets of parameters.
        """
        custom_param_groups = optimizer_opts.pop("param_groups_custom", None)
        if custom_param_groups is not None:
            assert callable(custom_param_groups), f"invalid param_groups_custom {custom_param_groups!r}"
            rf_model = wrapped_pt_module_to_rf_module(self.network)
            custom_param_groups = custom_param_groups(
                model=self.network, rf_model=rf_model, optimizer_class=optim_class, optimizer_opts=optimizer_opts
            )
            return custom_param_groups

        network_params = self.network.parameters()

        # By default, insert the weight_decay constraints in the optimizer, as this is default PyTorch behavior.
        # If the user doesn't accept this, throw an error message.
        assert self.config.bool("decouple_constraints", True), (
            "L2/weight_decay constraints are decoupled in PyTorch, but "
            "decouple_constraints=False was explicitly specified in the config."
        )

        # Split in parameter groups only if decouple_constraints is set and the optimizer accepts weight_decay.
        cls_init_kwargs = _get_class_init_kwargs(optim_class)
        if "weight_decay" not in cls_init_kwargs:
            assert "weight_decay" not in optimizer_opts, (
                "weight_decay not accepted by the chosen optimizer. Accepted values: %s"
                % ", ".join("%s" % optim_name for optim_name in cls_init_kwargs)
            )
            return network_params

        weight_decay = optimizer_opts.get("weight_decay", 0.0)
        if not weight_decay:
            return network_params

        # Distinguish between parameters with and without weight_decay/L2 regularization.
        # Parameters without weight decay: biases + LayerNorm/Embedding layers.
        wd_params = set()
        no_wd_params = set()
        blacklist_wd_modules = optimizer_opts.pop("weight_decay_modules_blacklist", None)
        if blacklist_wd_modules is None:
            blacklist_wd_modules = (torch.nn.LayerNorm, torch.nn.Embedding)
        else:
            blacklist_wd_modules = _wrap_user_blacklist_wd_modules(blacklist_wd_modules)
        custom_include_check = optimizer_opts.pop("weight_decay_custom_include_check", None)
        if custom_include_check:
            assert callable(custom_include_check), f"invalid weight_decay_custom_include_check {custom_include_check!r}"
        # Tracker of visited parameters to only add each parameter once, in case two modules share common parameters.
        # We need the wrapper class RefIdEq because Parameters are compared by value and not by reference.
        visited_params: Set[RefIdEq[torch.nn.Parameter]] = set()
        for module_name, module in self.network.named_modules():
            module_name: str
            module: torch.nn.Module
            rf_module = wrapped_pt_module_to_rf_module(module)
            for param_name, param in module.named_parameters(recurse=False):
                param_name: str
                param: torch.nn.Parameter
                if RefIdEq(param) in visited_params:
                    continue
                visited_params.add(RefIdEq(param))
                full_param_name = "%s.%s" % (module_name, param_name) if module_name else param_name
                custom_include = None
                if custom_include_check:
                    custom_include = custom_include_check(
                        module=module, rf_module=rf_module, full_param_name=param_name, param=param
                    )
                if custom_include is not None:
                    assert isinstance(custom_include, bool), "weight_decay_custom_include_check did not return bool"
                    if custom_include:
                        wd_params.add(full_param_name)
                    else:
                        no_wd_params.add(full_param_name)
                elif (
                    param_name.endswith("bias")
                    or isinstance(module, blacklist_wd_modules)
                    or isinstance(rf_module, blacklist_wd_modules)
                ):
                    no_wd_params.add(full_param_name)
                else:
                    wd_params.add(full_param_name)

        param_dict = {pn: p for pn, p in self.network.named_parameters()}
        optim_groups = [
            {"params": [param_dict[pn] for pn in sorted(list(wd_params))], "weight_decay": weight_decay},
            {"params": [param_dict[pn] for pn in sorted(list(no_wd_params))], "weight_decay": 0.0},
        ]

        return optim_groups


def _wrap_user_blacklist_wd_modules(
    mods: Sequence[Union[str, Type[rf.Module], Type[torch.nn.Module]]],
) -> Tuple[type, ...]:
    assert isinstance(mods, (list, tuple)), f"invalid blacklist_weight_decay_modules {mods!r}"
    res = []
    for mod in mods:
        if isinstance(mod, str):
            assert mod.startswith("torch.") or mod.startswith("rf."), f"invalid blacklist_weight_decay_modules {mods!r}"
            mod = eval(mod)
        assert issubclass(mod, (rf.Module, torch.nn.Module)), f"invalid blacklist_weight_decay_modules {mods!r}"
        res.append(mod)
    return tuple(res)


def gradient_noise_(params: Iterable[torch.nn.Parameter], std: float):
    """
    Add gradient noise to parameters, using a truncated normal distribution.
    """
    a, b = -2 * std, 2 * std
    for param in params:
        if param.requires_grad and param.grad is not None:
            noise = torch.empty_like(param.grad)
            torch.nn.init.trunc_normal_(noise, std=std, a=a, b=b)
            param.grad += noise



"""
RETURNN frontend (returnn.frontend) utils
"""

from __future__ import annotations
from typing import Optional, Union, Any, Dict, Tuple
import contextlib
import os
import re
import numpy
import numpy.testing

from returnn.config import Config, global_config_ctx
from returnn.util.pprint import pprint
import returnn.frontend as rf
from returnn.tensor import Tensor, Dim, TensorDict, batch_dim
from returnn.tensor.utils import tensor_dict_fill_random_numpy_
import returnn.torch.frontend as rft
from returnn.torch.data.tensor_utils import tensor_dict_numpy_to_torch_, tensor_dict_torch_to_numpy_
from returnn.torch.util.debug_inf_nan import debug_inf_nan as torch_debug_inf_nan

# noinspection PyProtectedMember
from returnn.frontend._random_journal import RandomJournal


disable_tf = "RETURNN_DISABLE_TF" in os.environ and int(os.environ["RETURNN_DISABLE_TF"]) == 1

if disable_tf:
    tf = None
else:
    try:
        import tensorflow as tf
    except ImportError:
        tf = globals().get("tf", None)  # type: ignore

if tf:
    import returnn.tf.compat as tf_compat
    import returnn.tf.frontend_layers as rfl
    from returnn.tf.network import TFNetwork
else:
    tf_compat = rfl = TFNetwork = None


@contextlib.contextmanager
def tf_scope():
    """tf scope"""
    with tf_compat.v1.Graph().as_default(), tf_compat.v1.Session().as_default() as session:
        yield session


def run_model(
    extern_data: TensorDict,
    get_model: rf.GetModelFunc,
    forward_step: rf.StepFunc,
    *,
    dyn_dim_max_sizes: Optional[Dict[Dim, int]] = None,
    dyn_dim_min_sizes: Optional[Dict[Dim, int]] = None,
    test_tensorflow: bool = True,
    allow_inf_nan_in_output: bool = False,
    test_single_batch_entry: bool = True,
) -> TensorDict:
    """run"""
    print(f"* run_model with dyn_dim_max_sizes={dyn_dim_max_sizes!r}")
    extern_data_dims = extern_data.all_dims()
    extern_data.reset_content()
    tensor_dict_fill_random_numpy_(
        extern_data, dyn_dim_max_sizes=dyn_dim_max_sizes, dyn_dim_min_sizes=dyn_dim_min_sizes
    )

    print("** run with PyTorch backend")
    with rft.TorchBackend.random_journal_record() as random_journal:
        out_pt = _run_model_torch(extern_data, get_model, forward_step)
        _pad_mask_zeros(out_pt)
        # get the values now because dims might get overwritten
        out_pt_raw = out_pt.as_raw_tensor_dict(include_const_sizes=True)

    if not allow_inf_nan_in_output:
        non_finite_outputs = {}
        for k, v in out_pt.data.items():
            if not numpy.isfinite(v.raw_tensor).all():
                non_finite_outputs[k] = v
                print(f"ERROR: output {k!r} has non-finite values:\n{v.raw_tensor}")
        if non_finite_outputs:
            torch_debug_inf_nan(
                lambda: (_run_model_torch(extern_data, get_model, forward_step), None)[-1],
                stop_reporting_after_first_inf_nan=False,
            )
            raise Exception(f"Non-finite values in output: {non_finite_outputs}. See log above.")

    if test_single_batch_entry and batch_dim in extern_data_dims:
        dyn_dims = [
            d
            for d in extern_data_dims
            if d.dyn_size_ext is not None
            and d.dyn_size_ext.dims == (batch_dim,)
            and d.dyn_size_ext.raw_tensor.min() != d.dyn_size_ext.raw_tensor.max()
        ]
        if dyn_dims:  # e.g. the typical time dim with dyn size shape [batch_dim]
            batch_idx = dyn_dims[0].dyn_size_ext.raw_tensor.argmin().item()
            print(f"** run with PyTorch backend with single batch entry for some smaller sequence {batch_idx=}")
            for d in dyn_dims:
                print(f"  {d}: {d.dyn_size_ext.raw_tensor}")
            _run_model_torch_single_batch(extern_data, get_model, forward_step, batch_idx=batch_idx, ref_output=out_pt)

    if not test_tensorflow:
        return out_pt
    if disable_tf:
        print("** TensorFlow disabled (RETURNN_DISABLE_TF)")
        return out_pt

    assert tf, "TensorFlow not available"
    print("** run with TensorFlow-net-dict backend")
    with rfl.ReturnnLayersBackend.random_journal_replay(random_journal):
        out_tf = _run_model_net_dict_tf(extern_data, get_model, forward_step)
        _pad_mask_zeros(out_tf)
        out_tf_raw = out_tf.as_raw_tensor_dict(include_const_sizes=True)

    random_journal: RandomJournal
    assert random_journal.reached_end()

    print("Output PT/TF:", out_pt, out_tf)
    assert set(out_pt.data.keys()) == set(out_tf.data.keys()), (
        f"PT output {list(out_pt.data.keys())} vs TF output {list(out_tf.data.keys())}"
    )
    for k, v_pt in out_pt.data.items():
        v_tf = out_tf[k]
        # We cannot really check the dims directly for equality,
        # because the model code often creates new dims, which are different in each call.
        # However, via mark_as_output, the order of dims is well-defined.
        # So we can check the values.
        assert len(v_pt.dims) == len(v_tf.dims)
        assert v_pt.feature_dim_axis == v_tf.feature_dim_axis
        for d_pt, d_tf in zip(v_pt.dims, v_tf.dims):
            _check_dim(d_pt, d_tf)
        if v_pt.dtype.startswith("int"):
            assert v_tf.dtype.startswith("int")  # allow maybe different bit depth
        else:
            assert v_pt.dtype == v_tf.dtype, f"PT dtype {v_pt.dtype} vs TF dtype {v_tf.dtype}"
        assert bool(v_pt.sparse_dim) == bool(v_tf.sparse_dim)
        if v_pt.sparse_dim:
            _check_dim(v_pt.sparse_dim, v_tf.sparse_dim)
    assert set(out_pt_raw.keys()) == set(out_tf_raw.keys())
    for k, v_pt in out_pt_raw.items():
        v_tf = out_tf_raw[k]
        print(f"  comparing {k!r} {_array_repr(v_pt)} PT vs TF")
        if not numpy.allclose(v_pt, v_tf, atol=1e-5, rtol=1e-5):
            print(f"  PT:\n{v_pt}")
            print(f"  TF:\n{v_tf}")
            raise Exception(f"output {k!r} differs")
    return out_pt


def _run_model_torch(extern_data: TensorDict, get_model: rf.GetModelFunc, forward_step: rf.StepFunc) -> TensorDict:
    """run"""
    extern_data_raw = extern_data.as_raw_tensor_dict(expected_value_type=numpy.ndarray)
    rf.select_backend_torch()
    rf.set_random_seed(42)

    # Inplace replace Numpy by Torch.
    # Inplace because dim tags cannot easily be copied (then they are not the same).
    # We recover extern_data in the end.
    tensor_dict_numpy_to_torch_(extern_data)

    for k, v in extern_data.data.items():
        if v.raw_tensor.dtype.is_floating_point:
            v.raw_tensor.requires_grad = True

    model = get_model(epoch=1, step=0)
    rf.init_forward_step_run_ctx(epoch=1, step=0)
    forward_step(model=model, extern_data=extern_data)
    outputs = rf.get_run_ctx().outputs
    assert outputs.data

    if "loss" in outputs.data:
        loss = outputs.data["loss"]
        assert isinstance(loss, Tensor)
        assert loss.raw_tensor.dtype.is_floating_point
        loss = rf.reduce_sum(loss, axis=loss.dims)
        print("loss:", loss.raw_tensor.detach().numpy().item())
        loss.raw_tensor.backward()
        for k, v in list(extern_data.data.items()):
            if v.raw_tensor.dtype.is_floating_point:
                assert v.raw_tensor.grad is not None, f"no grad for {k}"
                v_grad = v.copy_template()
                v_grad.raw_tensor = v.raw_tensor.grad
                assert f"{k}_grad" not in outputs.data
                outputs.data[f"{k}_grad"] = v_grad

    tensor_dict_torch_to_numpy_(outputs)

    extern_data.assign_from_raw_tensor_dict_(extern_data_raw)
    return outputs


def run_model_torch_train(
    extern_data: TensorDict,
    get_model: rf.GetModelFunc,
    train_step: rf.StepFunc,
    *,
    dyn_dim_max_sizes: Optional[Dict[Dim, int]] = None,
    dyn_dim_min_sizes: Optional[Dict[Dim, int]] = None,
) -> Dict[str, float]:
    """run"""
    rf.select_backend_torch()
    rf.set_random_seed(42)

    extern_data.reset_content()
    tensor_dict_fill_random_numpy_(
        extern_data, dyn_dim_max_sizes=dyn_dim_max_sizes, dyn_dim_min_sizes=dyn_dim_min_sizes
    )
    tensor_dict_numpy_to_torch_(extern_data)

    # We want to be able to calculate gradients for testing,
    # so we need to set requires_grad=True.
    for v in extern_data.data.values():
        v: Tensor
        v.raw_tensor.requires_grad = True

    model = get_model(epoch=1, step=0)
    rf.init_train_step_run_ctx(train_flag=True, step=0, epoch=1)
    train_step(model=model, extern_data=extern_data)
    total_loss = rf.get_run_ctx().total_loss()
    assert isinstance(total_loss, Tensor) and not total_loss.dims and total_loss.raw_tensor.dtype.is_floating_point
    total_loss_v = total_loss.raw_tensor.detach().numpy().item()
    print("total loss (for backprop):", total_loss_v)
    res = {"total_loss": total_loss_v}

    total_loss.raw_tensor.backward()  # test backprop

    for k, loss in rf.get_run_ctx().losses.items():
        loss_v = loss.get_summed_loss().raw_tensor.detach().cpu().numpy().item()
        res[f"{k}:summed"] = loss_v
        print(f"loss (summed) {k!r}: {loss_v}")
        loss_v = loss.get_mean_loss().raw_tensor.detach().cpu().numpy().item()
        print(f"loss (mean) {k!r}: {loss_v}")
        res[f"{k}:mean"] = loss_v
        inv_norm_factor = loss.get_inv_norm_factor()
        if isinstance(inv_norm_factor, Tensor):
            inv_norm_factor = inv_norm_factor.raw_tensor.detach().sum().cpu().numpy().item()
        print(f"inv_norm_factor {k!r}: {inv_norm_factor}")
        res[f"{k}:inv_norm_factor"] = inv_norm_factor

    return res


def _run_model_torch_single_batch(
    extern_data: TensorDict,
    get_model: rf.GetModelFunc,
    forward_step: rf.StepFunc,
    *,
    batch_idx: int,
    ref_output: TensorDict,
):
    """
    Restrict batch_dim to batch_idx only.
    This is somewhat hacky:
    We want to keep the same dim tags.
    Thus, we overwrite all the data by the sliced data.
    We also want to keep the batch_dim for the forward_step as the user code probably expects it.
    Afterward, we want to recover the original data.
    """
    # Store original data to be able to recover it later.
    extern_data_raw = extern_data.as_raw_tensor_dict(expected_value_type=numpy.ndarray)

    # noinspection PyShadowingNames
    def _get_slices(x: Tensor) -> Tuple[slice, ...]:
        slices = []
        for dim in x.dims:
            if dim == batch_dim:
                slices.append(slice(batch_idx, batch_idx + 1))
            elif dim.dyn_size_ext is not None and batch_dim in dim.dyn_size_ext.dims:
                slices.append(slice(0, dim.dyn_size_ext.raw_tensor.max().item()))
            else:
                slices.append(slice(None))
        return tuple(slices)

    # inplace
    # noinspection PyShadowingNames
    def tensor_numpy_restrict_batch_dim_(x: Tensor[numpy.ndarray]):
        if batch_dim not in x.dims:
            return
        if x.raw_tensor is not None:
            assert isinstance(x.raw_tensor, numpy.ndarray)
            x.raw_tensor = x.raw_tensor[_get_slices(x)]

    batch_dim.reset_eager()
    batch_dim.dyn_size_ext = Tensor("batch_size", dims=(), dtype="int32", raw_tensor=numpy.array(1, dtype="int32"))
    for dim in extern_data.all_dims():
        dim.transform_tensors(tensor_numpy_restrict_batch_dim_)
    for v in extern_data.data.values():
        tensor_numpy_restrict_batch_dim_(v)

    output = _run_model_torch(extern_data, get_model, forward_step)
    for key, ref_output_ in ref_output.data.items():
        output_ = output.data[key]
        if batch_dim not in ref_output_.dims:
            continue
        batch_axis = ref_output_.dims.index(batch_dim)
        assert output_.dims[batch_axis] == batch_dim and len(output_.dims) == len(ref_output_.dims)
        # Slice the raw ref output to be able to match it to the raw single output.
        ref_output_raw = ref_output_.raw_tensor[_get_slices(output_)]
        single_output_raw = output_.raw_tensor
        numpy.testing.assert_allclose(
            ref_output_raw, single_output_raw, atol=1e-5, rtol=1e-5, err_msg=f"output {key!r} differs"
        )

    # Recover original data.
    extern_data.reset_content()
    extern_data.assign_from_raw_tensor_dict_(extern_data_raw)


def _run_model_net_dict_tf(
    extern_data: TensorDict, get_model: rf.GetModelFunc, forward_step: rf.StepFunc
) -> TensorDict:
    """run"""
    extern_data_raw = extern_data.as_raw_tensor_dict(expected_value_type=numpy.ndarray)
    extern_data.reset_content()
    rf.select_backend_returnn_layers_tf()
    rf.set_random_seed(42)

    from returnn.tf.frontend_layers.config_entry_points import get_net_dict

    # noinspection PyProtectedMember
    from returnn.frontend import _backend

    config = Config(
        {
            "debug_runtime_sanity_checks": True,
            # "debug_print_layer_output": True,  # enable temporarily for debugging
            "extern_data": extern_data,
            "get_model": get_model,
            "task": "forward",
            "forward_step": forward_step,
        }
    )

    with tf_scope() as session, global_config_ctx(config):
        net_dict, model = get_net_dict(epoch=1, step=0)
        print("*** TF net dict:")
        pprint(net_dict)
        outputs_layers = rf.get_run_ctx().outputs
        print("*** outputs:", outputs_layers)

        _backend.select_backend_tf()
        net = TFNetwork(config=config, train_flag=False)
        net.construct_from_dict(net_dict)

        rf_params = {name.replace(".", "/"): p for name, p in model.named_parameters()}
        tf_params = {re.sub("/param:0$", "", p.name): p for p in net.get_params_list()}
        rf_params_not_in_tf = set(rf_params.keys()) - set(tf_params.keys())
        tf_params_not_in_rf = set(tf_params.keys()) - set(rf_params.keys())
        if rf_params_not_in_tf or tf_params_not_in_rf:
            raise Exception(
                "params not equal:\n"
                f"RF params not in TF: {rf_params_not_in_tf}\n"
                f"TF params not in RF: {tf_params_not_in_rf}"
            )

        session.run(tf_compat.v1.global_variables_initializer())

        outputs_tf = TensorDict()
        for k, v in outputs_layers.data.items():
            v: Tensor[rfl.Layer]
            assert isinstance(v.raw_tensor, rfl.Layer)
            layer_name = v.raw_tensor.get_abs_name()
            layer = net.get_layer(layer_name)
            outputs_tf.data[k] = layer.output.copy()

        if "loss" in outputs_tf.data:
            data_ = {name: data for name, data in net.extern_data.data.items() if data.dtype.startswith("float")}
            loss = outputs_tf.data["loss"]
            assert isinstance(loss, Tensor)
            assert loss.dtype.startswith("float")
            loss = rf.reduce_sum(loss, axis=loss.dims)
            d_grads = tf.gradients(loss.raw_tensor, [d.raw_tensor for d in data_.values()])
            for (name, data), d_grad_tf in zip(data_.items(), d_grads):
                assert isinstance(data, Tensor)
                assert isinstance(d_grad_tf, tf.Tensor)
                d_grad = data.copy_template()
                d_grad.raw_tensor = d_grad_tf
                outputs_tf.data[f"{name}_grad"] = d_grad

        fetches = outputs_tf.as_raw_tensor_dict(expected_value_type=tf.Tensor)
        assert set(extern_data.data.keys()) == set(net.extern_data.data.keys())
        extern_data_tf_placeholders = net.extern_data.as_raw_tensor_dict(expected_value_type=tf.Tensor)
        assert set(extern_data_tf_placeholders.keys()) == set(extern_data_raw.keys())
        feed_dict = {extern_data_tf_placeholders[k]: v for k, v in extern_data_raw.items()}

        outputs_numpy_raw = session.run(fetches, feed_dict=feed_dict)

        # Scalars are not Numpy arrays, but our code below assumes that we only have Numpy arrays.
        # So we convert them here.
        def _make_numpy_array(x):
            if isinstance(x, numpy.ndarray):
                return x
            return numpy.array(x)

        outputs_numpy_raw = {k: _make_numpy_array(v) for k, v in outputs_numpy_raw.items()}

        outputs_numpy = outputs_tf.copy_template()
        outputs_numpy.reset_content()
        outputs_numpy.assign_from_raw_tensor_dict_(outputs_numpy_raw)

    extern_data.assign_from_raw_tensor_dict_(extern_data_raw)
    return outputs_numpy


def _dim_is_scalar_size(dim: Dim) -> bool:
    """dim is scalar size"""
    if dim.size is not None:
        return True
    if dim.dyn_size_ext is not None:
        return dim.dyn_size_ext.dims == ()
    return False


def _dim_scalar_size(dim: Dim) -> int:
    """dim scalar size"""
    if dim.size is not None:
        return dim.size
    if dim.dyn_size_ext is not None:
        assert dim.dyn_size_ext.dims == ()
        return dim.dyn_size_ext.raw_tensor
    raise Exception(f"dim {dim} has no known size")


def _pad_mask_zeros(x: Union[TensorDict, Tensor, Dim]):
    if isinstance(x, TensorDict):
        for v in x.data.values():
            _pad_mask_zeros(v)
        return

    if isinstance(x, Dim):
        if x.dyn_size_ext is not None:
            _pad_mask_zeros(x.dyn_size_ext)
        return

    assert isinstance(x, Tensor)
    x.raw_tensor = x.copy_masked(0).raw_tensor
    for d in x.dims:
        _pad_mask_zeros(d)


def _check_dim(d_pt: Dim, d_tf: Dim):
    assert isinstance(d_pt, Dim) and isinstance(d_tf, Dim)
    assert d_pt.size == d_tf.size
    assert _dim_is_scalar_size(d_pt) == _dim_is_scalar_size(d_tf)
    if not _dim_is_scalar_size(d_pt):
        assert d_pt.dyn_size_ext is not None and d_tf.dyn_size_ext is not None
        # There might be cases where the dims are maybe not equal
        # (same reasoning as above, or also different order),
        # although this would be quite exotic.
        # Let's just assume for now that this does not happen.
        assert d_pt.dyn_size_ext.dims == d_tf.dyn_size_ext.dims
        # Do not compare the raw tensor directly here.
        # This will be done later for all the raw output.
        # The reason is that the different backends (TF, PT, etc)
        # might share the same dim tag but then reset them.


def _walk_dims(start: Dim, *, func=print):
    visited = set()  # ids
    queue = [((), start)]
    while queue:
        path, dim = queue.pop(0)
        path: Tuple[Any, ...]
        dim: Dim
        if id(dim) in visited:
            continue
        visited.add(id(dim))
        func(path, dim)
        # noinspection PyProtectedMember
        dim_extra = dim._extra
        if dim_extra:
            if dim_extra.cache_dim_math:
                for k, v in dim_extra.cache_dim_math.items():
                    k: Any
                    queue.append((path + ("_extra.cache_dim_math", k), v))
            if dim_extra.same_as:
                queue.append((path + ("_extra.same_as",), dim_extra.same_as))
            if dim_extra.derived_from_op:
                for i, v in enumerate(dim_extra.derived_from_op.inputs):
                    queue.append((path + ("_extra.derived_from_op.inputs", i), v))
            for k, v in dim_extra.same_for_batch_ctx.items():
                k: Any
                queue.append((path + ("_extra.same_for_batch_ctx", k), v))


def _array_repr(x: Union[numpy.ndarray, numpy.number]) -> str:
    if not isinstance(x, numpy.ndarray):
        return f"<{type(x).__name__} {x!r}>"

    try:
        import lovely_numpy

        return f"<{lovely_numpy.lovely(x)}>"
    except ImportError:
        if x.size <= 10:
            return repr(x)
        return f"<array shape={x.shape} dtype={x.dtype} min={x.min()} max={x.max()}>"



"""
DType helpers
"""

from __future__ import annotations
from contextlib import contextmanager


__all__ = [
    "get_default_float_dtype",
    "set_default_float_dtype",
    "set_default_float_dtype_ctx",
    "get_default_int_dtype",
    "get_default_array_index_dtype",
    "is_float_dtype",
]


_default_float_dtype: str = "float32"
_default_int_dtype: str = "int32"


def get_default_float_dtype() -> str:
    """
    https://data-apis.org/array-api/latest/API_specification/data_types.html#default-data-types

    :return: default dtype for float
    """
    return _default_float_dtype


def set_default_float_dtype(dtype: str):
    """
    Set the default float dtype

    :param dtype: the new default float dtype
    """
    global _default_float_dtype
    assert isinstance(dtype, str)
    _default_float_dtype = dtype


@contextmanager
def set_default_float_dtype_ctx(dtype: str):
    """
    :param dtype: see :func:`get_default_float_dtype`
    """
    global _default_float_dtype
    assert isinstance(dtype, str)
    old_default_float_dtype = _default_float_dtype
    try:
        _default_float_dtype = dtype
        yield
    finally:
        _default_float_dtype = old_default_float_dtype


def get_default_int_dtype() -> str:
    """
    https://data-apis.org/array-api/latest/API_specification/data_types.html#default-data-types

    :return: default dtype for int
    """
    return _default_int_dtype


def get_default_array_index_dtype() -> str:
    """
    https://data-apis.org/array-api/latest/API_specification/data_types.html#default-data-types

    :return: default dtype for array index - currently just the same as :func:`get_default_int_dtype`
    """
    return get_default_int_dtype()


def is_float_dtype(dtype: str) -> bool:
    """
    :return: whether the dtype is float, e.g. it supports backprop etc
    """
    return dtype.startswith("float")



#!/usr/bin/env python3

"""
This is a wrapper around :func:`returnn.tf.engine.Engine.cleanup_old_models`.
"""

from __future__ import annotations
import sys
import os
import argparse

import _setup_returnn_env  # noqa
from returnn.util import better_exchook
from returnn.log import log
from returnn.__main__ import init, finalize


def main():
    """
    Main entry.
    """
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument("--config")
    arg_parser.add_argument("--cwd", help="will change to this dir")
    arg_parser.add_argument("--model", help="model filenames (default: take from config)")
    arg_parser.add_argument("--scores", help="learning_rate_control file, e.g. newbob.data (default: take from config)")
    arg_parser.add_argument("--dry_run", action="store_true")
    args = arg_parser.parse_args()
    return_code = 0
    try:
        if args.cwd:
            os.chdir(args.cwd)
        init(
            extra_greeting="Delete old models.",
            config_filename=args.config or None,
            config_updates={"need_data": False, "device": "cpu"},
        )
        from returnn.__main__ import engine, config

        if args.model:
            config.set("model", args.model)
        if args.scores:
            config.set("learning_rate_file", args.scores)
        if args.dry_run:
            config.set("dry_run", True)
        engine.cleanup_old_models(ask_for_confirmation=True)

    except KeyboardInterrupt:
        return_code = 1
        print("KeyboardInterrupt", file=getattr(log, "v3", sys.stderr))
        if getattr(log, "verbose", [False] * 6)[5]:
            sys.excepthook(*sys.exc_info())
    finalize()
    if return_code:
        sys.exit(return_code)


if __name__ == "__main__":
    better_exchook.install()
    main()
